
-------------------------------------------------------------------------------
-- REAL-TIME PITCH DETECTION
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use IEEE.STD_LOGIC_ARITH.all;
use IEEE.STD_LOGIC_UNSIGNED.all;

entity pd is

  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector(0 to 31) := X"00000000";
    C_HIGHADDR   : std_logic_vector(0 to 31) := X"FFFFFFFF"
    );

  port (
    OPB_Clk     : in  std_logic;
    OPB_Rst     : in  std_logic;
    OPB_ABus    : in  std_logic_vector(0 to C_OPB_AWIDTH-1);  --(31:0)
    OPB_BE      : in  std_logic_vector(0 to C_OPB_DWIDTH/8-1);--(3:0)
    OPB_DBus    : in  std_logic_vector(0 to C_OPB_DWIDTH-1);  --(31:0)
    OPB_RNW     : in  std_logic;
    OPB_select  : in  std_logic;
    OPB_seqAddr : in  std_logic;        -- Sequential Address
    Sln_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);  --(31:0)
    Sln_errAck  : out std_logic;        -- (unused)
    Sln_retry   : out std_logic;        -- (unused)
    Sln_toutSup : out std_logic;        -- Timeout suppress
    Sln_xferAck : out std_logic;        -- Transfer acknowledge

    SRAM_CE     : out std_logic;        --sram chip enable (active low)
    SRAM_OE     : out std_logic;        --sram output enable (active low)
    SRAM_WE     : out std_logic;        --sram write enable (active low)
    SRAM_UB     : out std_logic;        --sram enable upper-byte(active low)
    SRAM_LB     : out std_logic;        --sram enable lower-byte(active low)
    PB_A        : out std_logic_vector(17 downto 0);  --sram 18 bit address
    PB_D        : inout std_ulogic_vector(15 downto 0);  --sram 16 bit data

   -- G: For audio codec --
    au_mclk     : out std_logic;
    au_lrclk    : out std_logic;
    au_bclk     : out std_logic;
    au_sdti     : out std_logic;
    au_sdto0    : in std_logic;
    au_cs       : out std_logic
    );
end pd;

-------------------------------------------------------------------------------

architecture Behavioral of pd is


-----------------------------------------------------------------------------
-- SRAM
-----------------------------------------------------------------------------
  
  constant RAM_AWIDTH : integer := 18;  -- Number of address lines on the RAM
  constant RAM_DWIDTH : integer := 16;  -- Number of data lines on the RAM
  
  component OBUF_F_24
    port (
      O : out STD_ULOGIC;               -- the pin
      I : in  STD_ULOGIC);              -- signal to pin
  end component;

  component IOBUF_F_24
    port (
      O : out STD_ULOGIC;               -- signal from pin
      IO : inout STD_ULOGIC;            -- the pin
      I : in  STD_ULOGIC;               -- signal to pin
      T : in STD_ULOGIC);               -- 1-drive IO with I
  end component;

-------------------------------------------------------------------------------
-- Audio Controller
-------------------------------------------------------------------------------
  
  component wr_audio_control
    port (
      clk_in: in std_logic;               -->200ns clock
      init: in std_logic;                 --pulse input
      done: out std_logic;
      D_in: in std_logic_vector(15 downto 0);        --control data
      cntrl_out: out std_logic);
  end component;


 ------------------------------------------------------------------------------
 -- Codec
 ------------------------------------------------------------------------------
  component ak4565
    port (
      clk      : in  std_logic;
      rst      : in  std_logic;
      mclk     : out std_logic;
      bclk     : out std_logic;
      lrclk    : out std_logic;
      sdti     : out std_logic;
      sdto0    : in  std_logic;
      csn      : out std_logic;
      cclk     : out std_logic;
      cdti     : out std_logic;

      adcdone : out std_logic;  --AD indicator
      dacload : out std_logic;  --DA indicator	--G: UNUSED
      adc_dtout : out std_logic_vector(15 downto 0);  --data out to fpga 
      dac_dtin  : in  std_logic_vector(15 downto 0);  --parallelload data from fpga
      c_datain : in std_logic_vector(15 downto 0);  --parallelload control from fpga
      c_wr     : in std_logic :='0';  --write to control
      c_done   : out std_logic  --write control done
      );
  end component;

-------------------------------------------------------------------------------
-- SIGNALS
-------------------------------------------------------------------------------
  
 
  signal mclk : std_logic;  --12.5Mhz
  signal bclk : std_logic;  --3.125Mhz
  signal lrclk: std_logic;  --48.8Khz
  signal cclk: std_logic;  

  signal sdti: std_logic;
  signal latch16bit: std_logic_vector(15 downto 0) := x"0000";

  signal adcdone : std_logic;
  signal dacload : std_logic;  --G: UNUSED
  signal adc_dtout : std_logic_vector(15 downto 0);  --data out to fpga 
  signal dac_dtin  : std_logic_vector(15 downto 0) :=x"0000";  --parallelload data from fpga --G: UNUSED
  signal c_datain : std_logic_vector(15 downto 0) :="1110000011100111";  --parallelload control from fpga
  signal csn : std_logic;
  signal c_wr     : std_logic;
  signal c_done   : std_logic;

  signal trigger : std_logic;

-------------------------------------------------------------------------------
  --SRAM

  signal tri_state : std_logic;                         --tristate
  signal ABus: std_logic_vector(17 downto 0);           --Abus --always equal to sram_addr
  signal pbDIn: std_logic_vector(15 downto 0);                  --RAM_DI
                                                                --is equal to sram_dt_wr when c_state=norm
                                                                --else is all zeroes
  
  signal sram_rnw: std_logic;                                   --RNW
  signal sram_addr: std_logic_vector(17 downto 0);              --Abus
  signal sram_dt_rd: std_logic_vector(15 downto 0);             --RAM_DO
  
  signal sram_dt_wr: std_logic_vector(15 downto 0);  --connects to pbDIn when c_state=norm

-- G added:
  signal chip_select : std_logic;
  signal output_enable, write_enable : std_logic;
  signal upperbyte_enable, lowerbyte_enable : std_logic;

  signal RNW : std_logic;
  signal RST : std_logic;
  
-- G added: some more stuff (for SRAM to be OPB readable (?))--

  -- Critical: Sln_xferAck is generated directly from state bit 0!
  constant STATE_BITS : integer :=3;
  constant Idle     : std_logic_vector(0 to STATE_BITS-1) := "000";
  constant Selected : std_logic_vector(0 to STATE_BITS-1) := "001";
  constant Read     : std_logic_vector(0 to STATE_BITS-1) := "011";
  constant Xfer     : std_logic_vector(0 to STATE_BITS-1) := "111";
  
  signal present_state, next_state : std_logic_vector(0 to STATE_BITS-1);

-----------------------------------------------------------------------------
  signal initcnt:std_logic_vector(15 downto 0):=X"0000";  --has to wait > 4128/fs=90ms for initialization
  constant pre_init : std_logic_vector(1 downto 0):="00";
  constant c_wr_wait: std_logic_vector(1 downto 0):="01";
  constant wait_done: std_logic_vector(1 downto 0):="10";
  constant norm : std_logic_vector(1 downto 0):="11";
  signal c_state, n_state :  std_logic_vector(1 downto 0):="00";  --used in both FSMs
  signal c_serial_data: std_logic;

  --used for FSM that controls writing audio data to memorys
  constant idling : std_logic_vector(1 downto 0):="00";
  constant s1: std_logic_vector(1 downto 0):="01";
  constant s2 : std_logic_vector(1 downto 0):="10";

  signal s_done : std_logic;            --when done writing to SRAM

------------------------------------------------------------------------------
--G: added:  (THESE SIGNALS WILL HELP COORDINATE WHO IS READING AND WRITING, but they are not yet used fully)
  signal audio_wr_addr: std_logic_vector(17 downto 0):="000000000000000000"; --the pointer to where the
                                                                             --current audio sample
                                                                             --should be written to in SRAM
  signal en_MB_read: std_logic;         --used to disable the microblaze from
                                        --reading when the data from codec is
                                        --being written to SRAM.
  signal MB_read_addr: std_logic_vector(17 downto 0);    --the current address where the
                                                         --Microblaze reads from.
  
-------------------------------------------------------------------------------

begin
------Audio Interface Signals and Buffering--------------------------------------------------

  au_mclk <= mclk;  -----VERY IMPORTANT---- mclk must be sync with LRCLK
  au_bclk <= bclk;
  au_lrclk<= lrclk;
  au_sdti <= sdti;
  au_cs <=csn;

------SRAM pin assignment-----------------------------------------------------------------------

  SRAM_CE <= '0' when chip_select = '1' else '1'; -- '0' when c_state=norm else1?
  SRAM_WE <= '0' when write_enable = '1' else '1';
  SRAM_OE <= '0' when output_enable = '1' else '1';  --sram_rnw?
  SRAM_UB <= '0' when upperbyte_enable = '1' else '1';
  SRAM_LB <= '0' when lowerbyte_enable = '1' else '1';  
    
  gen1: for m in 0 to 17 generate
    sramAddrpin:OBUF_F_24 port map (
      O=>PB_A(m),
      I=>ABus(m));
  end generate;

  gen2: for m in 0 to 15 generate
    sramDatapin:IOBUF_F_24 port map (
      O=>sram_dt_rd(m),
      IO=>PB_D(m),
      I=>pbDIn(m),
      T=>tri_state);
  end generate;

  
--G: we have to change this:----------------------------------------
  tri_state <=sram_rnw when c_state=norm else '0'; --always writing
  ABus<=sram_addr(17 downto 0);
  pbDIn(15 downto 2) <= sram_dt_wr(15 downto 2) when c_state=norm else (others => '0');
  pbDIn(0)<=sram_dt_wr(0) when c_state=norm else cclk;
  pbDIn(1)<=sram_dt_wr(1) when c_state=norm else c_serial_data;
--------------------------------------------------------------------

  chip_select <=
    '1' when OPB_select = '1' and
        OPB_ABus(0 to C_OPB_AWIDTH-2-RAM_AWIDTH) =
        C_BASEADDR(0 to C_OPB_AWIDTH-2-RAM_AWIDTH) else
    '0';
-------------------------------------------------------------------------------

  --The code below is from Lab6 and is for making the Microblaze be able to
  --read from the SRAM.  This way we can get the incoming audio data into the C-
  --code and do the signal processing in C.  It seems that although C is much
  --slower than doing it in hardware, there is enough speed to keep up real-
  --time with much decimation (i.e. skipping output points).  This just means a
  --less smooth output.
  
  -- Sequential part of the FSM
  fsm_seq : process(OPB_Clk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      present_state <= Idle;
    elsif OPB_Clk'event and OPB_Clk = '1' then
      present_state <= next_state;
    end if;
  end process fsm_seq;

  -- Combinational part of the FSM
  fsm_comb : process(OPB_Rst, present_state, chip_select, OPB_Select, RNW)
  begin
    RST <= '1';

    write_enable <= '0';
    output_enable <= '0';
    upperbyte_enable <= '1';
    lowerbyte_enable <= '1';
    tri_state <= '0';                  -- used to be "tristate"
    
    if OPB_RST = '1' then
      next_state <= Idle;
    else
      case present_state is
        when Idle =>
          if chip_select = '1' then
            next_state <= Selected;
          else
            next_state <= Idle;
          end if;

        when Selected =>
          
          if OPB_Select = '1' then
            if RNW = '1' then
              RST <= '0';
              next_state <= Read;
            else
              write_enable <= '1';
              tri_state <= '0';
              next_state <= Xfer;
            end if;
            
          else
            next_state <= Idle;
            
          end if;

        when Read =>
          if OPB_Select = '1' then
            output_enable <= '1';
            tri_state <= '1';
            next_state <= Idle;
          end if;

          -- State encoding is critical here: xfer must only be true here
        when Xfer =>
          next_state <= Idle;
          
        when others =>
          next_state <= Idle;
      end case;
    end if;
  end process fsm_comb;
  
  Sln_xferAck <= present_state(0);
  
  
-------------------------------------------------------------------------------
-- Audio codec port map
-------------------------------------------------------------------------------

  ak:ak4565 port map(clk =>OPB_Clk,
                     rst      =>OPB_Rst,
                     mclk     =>mclk,
                     bclk     =>bclk,
                     lrclk    =>lrclk,
                     sdti     =>sdti,
                     sdto0    =>au_sdto0,
                     csn      =>csn,
                     cclk     =>cclk,
                     cdti     =>c_serial_data,
                     adcdone =>adcdone,
                     dacload =>dacload,
                     adc_dtout => adc_dtout,
                     dac_dtin  => dac_dtin,
                     c_datain =>c_datain,
                     c_wr    =>c_wr,
                     c_done  =>c_done);
  
-------------------------------------------------------------------------------
-- PROCESSES
-------------------------------------------------------------------------------
  
  ---latch output data---------------------- (from codec)
  process (OPB_Clk,OPB_Rst)
  begin
    if OPB_Rst = '1' then
      latch16bit<=x"0000";
    elsif OPB_clk'event and OPB_clk ='1' then
      if adcdone = '1' then     --and dacload = '1' then  --G: don't need
        latch16bit(15 downto 0)<=adc_dtout(15 downto 0);
      end if;
    end if;
  end process;
  
-------------------------------------------------------------------------------
--G:added 
 
--Write the latched data to the SRAM, but divide by 2 first.
  process(OPB_Clk, OPB_Rst)
    begin
      if OPB_Rst='1' then
      sram_dt_wr(15)<=latch16bit(15);
      sram_dt_wr(14)<=latch16bit(15);
      sram_dt_wr(13)<=latch16bit(14);
      sram_dt_wr(12)<=latch16bit(13);
      sram_dt_wr(11)<=latch16bit(12);
      sram_dt_wr(10)<=latch16bit(11);
      sram_dt_wr(9)<=latch16bit(10);
      sram_dt_wr(8)<=latch16bit(9);
      sram_dt_wr(7)<=latch16bit(8);
      sram_dt_wr(6)<=latch16bit(7);
      sram_dt_wr(5)<=latch16bit(6);
      sram_dt_wr(4)<=latch16bit(5);
      sram_dt_wr(3)<=latch16bit(4);
      sram_dt_wr(2)<=latch16bit(3);
      sram_dt_wr(1)<=latch16bit(2);
      sram_dt_wr(0)<=latch16bit(1);
      --sram_dt_wr(14 downto 0)<=latch16bit(15 downto 1);
    end if;
  end process;

  --fsm of delay process-----
  process(OPB_Clk,OPB_Rst)
  begin
    if OPB_Rst='1' then
      c_state<=idling;
    elsif OPB_Clk'event and OPB_Clk='1' then
      c_state<=n_state;
    end if;
  end process;

  --state machine combinational logic
  FSM_Comb_wr_to_SRAM: process(OPB_Rst,trigger,c_state)
  begin
    sram_rnw<='1';  --always reading
    s_done<='0';

    if rst='1' then
      n_state <= idling;
    else
      case c_state is
        when idling =>
          if trigger='1' then
            n_state<=s1;
          else
            n_state<=idling;
          end if;
        when s1 =>
          --wait for calculation to be done
          n_state<=s2;
        when s2 =>
          --store new data to SRAM
          sram_rnw<='0';  --write to sram
          s_done<='1';
          n_state<=idling;
        when others =>
          n_state<=idling;
      end case;
    end if;
  end process;

--generate current address for writing or MB_reading.
  process(OPB_Clk,OPB_Rst)
  begin
    if rst='1' then
      sram_addr<="000000000000000000"; --sr_addr<="000101110111000000"; --starts at 24000
    elsif OPB_Clk'event and OPB_Clk='1' then
      if c_state=s1 then
        sram_addr<=audio_wr_addr;
      else
        sram_addr<=MB_read_addr;   --GUY: NO NEED (?)
      end if;
    end if;
  end process;

  --G: NOTE c_addr was changed to audio_wr_addr;
  
--update current SRAM address --
  process(OPB_Clk,OPB_Rst)
  begin
    if OPB_Rst='1' then
      audio_wr_addr <="000000000000000000";  --24000, 1/2sec delay
    elsif OPB_Clk'event and OPB_Clk='1' then
      audio_wr_addr <= audio_wr_addr + 1;
    end if;
  end process;

-------------------------------------------------------------------------------
  
--GUY:
  --- generate trigger for writing into SRAM when audio data is ready----------
  -- disables MicroBlaze from reading while data is being written to SRAM------
  process (OPB_Clk,OPB_Rst)
  begin
    if OPB_Rst = '1' then
      trigger<='0';
      en_MB_read <= '0';
    elsif OPB_clk'event and OPB_clk ='1' then
      if adcdone = '1' then
        trigger<='1';
        en_MB_read <= '0';
      else
        trigger<='0';
        en_MB_read <= '1';
      end if;
    end if;
  end process;


  ----control initialization for audio codec-------
  --op0 op1 op2 a0 a1 a2 a3 a4 d0 d1 d2 d3 d4 d5 d6 d7
  c_datain<= "1110000000100000";  --c_datain<= "1110100010001000";

  process (bclk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      c_state <= pre_init;
    elsif bclk'event and bclk='1' then
      c_state<=n_state;
    end if;
  end process;

  process (lrclk, OPB_Rst)
  begin
    if OPB_Rst = '1' then
      initcnt <= X"0000";
    elsif lrclk'event and lrclk='1' then
      initcnt <= initcnt + 1;
    end if;
  end process;

  process (initcnt(15),c_state,c_done)
  begin
    c_wr<='0';
    case c_state is
      when pre_init =>
        --wait for initcnt
        if initcnt(15) = '1' then
          n_state<=c_wr_wait;
        else
          n_state<=pre_init;
        end if;
      when c_wr_wait =>
        --hold c_wr for 1 cycle
        c_wr<='1';
        n_state<=wait_done;
      when wait_done =>
        if c_done='1' then
          --normal operation
          n_state<=norm;
        else
          n_state<=wait_done;
        end if;
      when norm =>
        n_state<=norm;

      when others =>
        n_state<=pre_init;
    end case;
  end process;

end Behavioral;
